(*** Règles du jeu ***)

(* Deux joueurs s'affrontent : les ronds et les croix.*)
type joueur = Rond | Croix ;;

(* Une case peut être vide ou occupée par le jeton d'un joueur. *)
type case = Jeton of joueur | Vide ;;

(* Une partie se joue sur neuf cases. *)
type partie = case list ;;

(* Une position est un entier.
 1 | 2 | 3 
---+---+---
 4 | 5 | 6 
---+---+---
 7 | 8 | 9 
*)
type position = int * int ;;

let pair2pos pair = match pair with (l,c) -> (3*c + l) ;;
let pos2pair pos = (pos/3 , pos mod 3) ;;

(* Nouvelle partie = 9 cases vides. *)
let nouveau_jeu = [
	Vide ; Vide ; Vide ;
	Vide ; Vide ; Vide ;
	Vide ; Vide ; Vide
] ;;

(* Un jeu doit être une liste de neuf cases. *)
exception JeuInvalide ;;

(* To string. *)
let joueur2str p =
	match p with
	  Croix -> "Joueur X"
	| Rond  -> "Joueur O"
;;

let case2str c =
	match c with
	  Jeton(Croix) -> "X"
	| Jeton(Rond)  -> "O"
	| Vide  -> "."
;;

let jeu2str p =
	match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] ->
		(String.concat " " [ "" ;
			(case2str hg) ; (case2str hc) ; (case2str hd) ; "\n" ;
			(case2str mg) ; (case2str mc) ; (case2str md) ; "\n" ;
			(case2str bg) ; (case2str bc) ; (case2str bd) ; "\n"
		])
	| _ -> raise JeuInvalide
;;

(* Les rangées gagnantes sont des séries de trois cases :
	3 rangées en ligne, 3 rangées en colonne, et 2 rangées en diagonale.
	
	FIXME: Ce code est factorisable, mais bon... *)
let lh p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ hg ; hc ; hd ]
	| _ -> raise JeuInvalide ;;

let lm p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ mg ; mc ; md ]
	| _ -> raise JeuInvalide ;;

let lb p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ bg ; bc ; bd ]
	| _ -> raise JeuInvalide ;;


let cg p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ hg ; mg ; bg ]
	| _ -> raise JeuInvalide ;;

let cc p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ hc ; mc ; bc ]
	| _ -> raise JeuInvalide ;;

let cd p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ hd ; md ; bd ]
	| _ -> raise JeuInvalide ;;


let dg p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ hg ; mc ; bd ]
	| _ -> raise JeuInvalide ;;

let dd p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] -> [ hd ; mc ; bg ]
	| _ -> raise JeuInvalide ;;

(* Un jeu est gagné si l'une des rangées gagnantes est remplie par un joueur. *)
let jeu_gagne p j =
	let win = [ Jeton(j) ; Jeton(j) ; Jeton(j) ] in
	if (lh p) = win || (lm p) = win || (lb p) = win
	|| (lh p) = win || (lm p) = win || (lb p) = win
	|| (dg p) = win || (dd p) = win
	then true
	else false
;;

let jeu_croix p = jeu_gagne p Croix ;;
let jeu_rond  p = jeu_gagne p Rond  ;;

(* Un jeu est jouable s'il reste une case vide. *)
let jouable p = match p with
	[ hg ; hc ; hd ; mg ; mc ; md ; bg ; bc ; bd ] ->
	 ( hg = Vide || hc = Vide || hd = Vide
	|| mg = Vide || mc = Vide || md = Vide
	|| bg = Vide || bc = Vide || bd = Vide )
	| _ -> raise JeuInvalide

	(* Un jeu est nul s'il n'est pas jouable ni gagné par un joueur. *)
let jeu_nul p =
	if (jouable p) || (jeu_croix p) || (jeu_rond p)
	then false
	else true
;;

(* Un jeu est fini s'il est gagné ou nul. *)
let jeu_fini p =
	(jeu_croix p) || (jeu_rond p) || (jeu_nul p)
;;

(* La position ne correspond pas à une case jouable. *)
exception PositionInvalide ;;

(* La case est déjà occupée. *)
exception CaseNonLibre ;;

(* Remplace le n-ième élément de liste par valeur. *)
let rec remplacer_si_vide liste valeur n =
	match liste with
	elem :: reste ->
		if n != 0 then
			elem :: (remplacer_si_vide reste valeur (n-1))
		else if elem = Vide then
				valeur :: reste
			else
				raise CaseNonLibre
	| [] -> raise PositionInvalide
;;

(* Retourne la partie p après que le joueur j ait joué en pos. *)
let jouer p j pos =
	remplacer_si_vide p (Jeton(j)) pos
;;

(* Retourne l'adversaire du joueur j. *)
let adversaire j =
	if j = Croix then Rond else Croix
;;

(*** Solveur ***)

(* Une configuration est l'état d'une partie au tour d'un joueur. *)
type config = Config of partie * joueur ;;

(* Un avenir lie une configuration à l'ensemble de ses avenirs possibles. *)
type avenir = Avenir of config * avenir list ;;

(* On construit la liste des cases libres en regardant chaque case (de 0 à 8).
Si la case qu'on regarde est vide, alors on note son numéro, sinon rien. *)
let rec lister_cases_libres p pos =
	if pos = 9 then []
	else let fin_de_liste = (lister_cases_libres p (pos+1)) in
	if (List.nth p pos) = Vide then
		pos :: fin_de_liste
	else
		fin_de_liste
	
;;

(* Prédire l'avenir de p pour j consiste à envisager tous les coups possibles
...puis à prédire l'avenir après chaque coup possible, et ainsi de suite. *)
let rec predire_avenir p j =
	let coups_possibles = lister_cases_libres p 0 in
	Avenir(Config(p,j) , predire_avenir_de_chaque_coup p j coups_possibles) 
and predire_avenir_de_chaque_coup p j lc =
	match lc with
	case :: reste ->
		let nj = (adversaire j)
		and np = (jouer p j case) in
			(predire_avenir np nj) :: (predire_avenir_de_chaque_coup p j reste)
	| [] -> []
;;

(* L'issue d'une partie est la victoire d'un joueur ou un match nul. *)
type issue = Victoire of joueur | Nul ;;

(* La meilleure de deux issues pour un joueur est j sa propre victoire,
suivie à défaut d'un match nul, suivie d'une victoire de l'adversaire. *)
let meilleure_issue i1 i2 j =
	if i1 = Victoire(j) || i2 = Victoire(j) then Victoire(j)
	else if i1 = Nul || i2 = Nul then Nul
	else Victoire(adversaire j)
;;

(* Déterminer l'issue d'une partie à partir d'un avenir se fait récursivement
par alternation : l'issue d'une partie terminée p est immédiate, l'issue
d'une partie p lorsqu'un joueur j s'apprête à jouer est la meilleure issue
possible pour ce joueur parmi la liste al des avenirs accessibles. *)
let rec determiner_issue a =
	match a with Avenir(Config(p,j), al) ->
		if (jeu_nul p) then Nul 
		else if (jeu_rond p) then Victoire(Rond)
		else if (jeu_croix p) then Victoire(Croix)
		else
			determiner_meilleure_issue al j
and determiner_meilleure_issue al j =
	match al with
	a :: reste ->
		 meilleure_issue (determiner_issue a) (determiner_meilleure_issue reste j) j
	| [] ->
		Victoire(adversaire j)
;;
		

(*** Tests ***)

let p0 = nouveau_jeu ;;
let p1 = (jouer p0 Rond  (pair2pos (0,0))) ;;
let p2 = (jouer p1 Croix (pair2pos (1,1))) ;;
let p3 = (jouer p2 Rond  (pair2pos (1,0))) ;;
let p4 = (jouer p3 Croix (pair2pos (2,0))) ;;
let p5 = (jouer p4 Rond  (pair2pos (0,2))) ;;

let game = p0 ;;
let joueur = Croix ;;

print_string (jeu2str game) ;
print_string "\nC'est parti..."

let arbre_des_possibles = (predire_avenir game joueur) ;;

print_string " done.\n"

let prediction = match (determiner_issue arbre_des_possibles) with
	Victoire(j) -> (joueur2str j)
	| Nul -> "Match nul."
;;

print_string "Prédiction : " ;
print_string prediction ;
print_string "\n"
